@@ -1,5 +1,5 @@
 // Copyright (c) 2020, Control Command Inc. All rights reserved.
-// Copyright (c) 2018-2022, Sylabs Inc. All rights reserved.
+// Copyright (c) 2018-2023, Sylabs Inc. All rights reserved.
 // This software is licensed under a 3-clause BSD license. Please consult the
 // LICENSE.md file distributed with the sources of this project regarding your
 // rights to use or distribute this software.
@@ -19,21 +19,25 @@ import (
 	"strings"
 	"text/template"
 
-	ocitypes "github.com/containers/image/v5/types"
+	"github.com/google/go-containerregistry/pkg/authn"
+	ggcrv1 "github.com/google/go-containerregistry/pkg/v1"
 	"github.com/spf13/cobra"
 	scskeyclient "github.com/sylabs/scs-key-client/client"
 	scslibclient "github.com/sylabs/scs-library-client/client"
-	"github.com/sylabs/singularity/docs"
-	"github.com/sylabs/singularity/internal/pkg/buildcfg"
-	"github.com/sylabs/singularity/internal/pkg/plugin"
-	"github.com/sylabs/singularity/internal/pkg/remote"
-	"github.com/sylabs/singularity/internal/pkg/remote/endpoint"
-	"github.com/sylabs/singularity/internal/pkg/util/fs"
-	"github.com/sylabs/singularity/pkg/cmdline"
-	clicallback "github.com/sylabs/singularity/pkg/plugin/callback/cli"
-	"github.com/sylabs/singularity/pkg/syfs"
-	"github.com/sylabs/singularity/pkg/sylog"
-	"github.com/sylabs/singularity/pkg/util/singularityconf"
+	"github.com/sylabs/singularity/v4/docs"
+	"github.com/sylabs/singularity/v4/internal/pkg/buildcfg"
+	"github.com/sylabs/singularity/v4/internal/pkg/ociplatform"
+	"github.com/sylabs/singularity/v4/internal/pkg/plugin"
+	"github.com/sylabs/singularity/v4/internal/pkg/remote"
+	"github.com/sylabs/singularity/v4/internal/pkg/remote/endpoint"
+	ocilauncher "github.com/sylabs/singularity/v4/internal/pkg/runtime/launcher/oci"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/fs"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/rootless"
+	"github.com/sylabs/singularity/v4/pkg/cmdline"
+	clicallback "github.com/sylabs/singularity/v4/pkg/plugin/callback/cli"
+	"github.com/sylabs/singularity/v4/pkg/syfs"
+	"github.com/sylabs/singularity/v4/pkg/sylog"
+	"github.com/sylabs/singularity/v4/pkg/util/singularityconf"
 	"golang.org/x/term"
 )
 
@@ -47,22 +51,11 @@ var CurrentUser = getCurrentUser()
 // currentRemoteEndpoint holds the current remote endpoint
 var currentRemoteEndpoint *endpoint.Config
 
-var (
-	dockerAuthConfig ocitypes.DockerAuthConfig
-	dockerLogin      bool
-
-	encryptionPEMPath   string
-	promptForPassphrase bool
-	forceOverwrite      bool
-	noHTTPS             bool
-	tmpDir              string
-)
-
 const (
 	envPrefix = "SINGULARITY_"
 )
 
-// singularity command flags
+// Top level options on the `singularity` root command.
 var (
 	debug   bool
 	nocolor bool
@@ -73,6 +66,46 @@ var (
 	configurationFile string
 )
 
+// Common options used with multiple sub-commands.
+var (
+	// OCI Registry Authentication
+	authConfig  authn.AuthConfig
+	dockerLogin bool
+	dockerHost  string
+	noHTTPS     bool
+
+	// Encryption Material
+	encryptionPEMPath   string
+	promptForPassphrase bool
+
+	// Paths / file handling
+	tmpDir         string
+	forceOverwrite bool
+
+	// Options controlling the unpacking of images to temporary sandboxes
+	canUseTmpSandbox bool
+	tmpSandbox       bool
+	noTmpSandbox     bool
+
+	// Use OCI runtime and OCI SIF?
+	isOCI bool
+	noOCI bool
+
+	// Keep individual layers when creating / pulling an OCI-SIF?
+	keepLayers bool
+
+	// Platform for retrieving images
+	arch     string
+	platform string
+
+	// Optional user requested authentication file for writing/reading OCI registry credentials
+	reqAuthFile string
+)
+
+//
+// Top level option flags
+//
+
 // -d|--debug
 var singDebugFlag = cmdline.Flag{
 	ID:           "singDebugFlag",
@@ -81,6 +114,7 @@ var singDebugFlag = cmdline.Flag{
 	Name:         "debug",
 	ShortHand:    "d",
 	Usage:        "print debugging information (highest verbosity)",
+	EnvKeys:      []string{"DEBUG"},
 }
 
 // --nocolor
@@ -122,26 +156,43 @@ var singVerboseFlag = cmdline.Flag{
 	Usage:        "print additional information",
 }
 
+// -c|--config
+var singConfigFileFlag = cmdline.Flag{
+	ID:           "singConfigFileFlag",
+	Value:        &configurationFile,
+	DefaultValue: buildcfg.SINGULARITY_CONF_FILE,
+	Name:         "config",
+	ShortHand:    "c",
+	Usage:        "specify a configuration file (for root or unprivileged installation only)",
+	EnvKeys:      []string{"CONFIG_FILE"},
+}
+
+//
+// Common option flags for multiple subcommands
+//
+
 // --docker-username
 var dockerUsernameFlag = cmdline.Flag{
-	ID:           "dockerUsernameFlag",
-	Value:        &dockerAuthConfig.Username,
-	DefaultValue: "",
-	Name:         "docker-username",
-	Usage:        "specify a username for docker authentication",
-	Hidden:       true,
-	EnvKeys:      []string{"DOCKER_USERNAME"},
+	ID:            "dockerUsernameFlag",
+	Value:         &authConfig.Username,
+	DefaultValue:  "",
+	Name:          "docker-username",
+	Usage:         "specify a username for docker authentication",
+	Hidden:        true,
+	EnvKeys:       []string{"DOCKER_USERNAME"},
+	WithoutPrefix: true,
 }
 
 // --docker-password
 var dockerPasswordFlag = cmdline.Flag{
-	ID:           "dockerPasswordFlag",
-	Value:        &dockerAuthConfig.Password,
-	DefaultValue: "",
-	Name:         "docker-password",
-	Usage:        "specify a password for docker authentication",
-	Hidden:       true,
-	EnvKeys:      []string{"DOCKER_PASSWORD"},
+	ID:            "dockerPasswordFlag",
+	Value:         &authConfig.Password,
+	DefaultValue:  "",
+	Name:          "docker-password",
+	Usage:         "specify a password for docker authentication",
+	Hidden:        true,
+	EnvKeys:       []string{"DOCKER_PASSWORD"},
+	WithoutPrefix: true,
 }
 
 // --docker-login
@@ -154,6 +205,37 @@ var dockerLoginFlag = cmdline.Flag{
 	EnvKeys:      []string{"DOCKER_LOGIN"},
 }
 
+// --docker-host
+var dockerHostFlag = cmdline.Flag{
+	ID:            "dockerHostFlag",
+	Value:         &dockerHost,
+	DefaultValue:  "",
+	Name:          "docker-host",
+	Usage:         "specify a custom Docker daemon host",
+	EnvKeys:       []string{"DOCKER_HOST"},
+	WithoutPrefix: true,
+}
+
+// --no-https
+var commonNoHTTPSFlag = cmdline.Flag{
+	ID:           "commonNoHTTPSFlag",
+	Value:        &noHTTPS,
+	DefaultValue: false,
+	Name:         "no-https",
+	Usage:        "use http instead of https for docker:// oras:// and library://<hostname>/... URIs",
+	EnvKeys:      []string{"NOHTTPS", "NO_HTTPS"},
+}
+
+// --nohttps (deprecated)
+var commonOldNoHTTPSFlag = cmdline.Flag{
+	ID:           "commonOldNoHTTPSFlag",
+	Value:        &noHTTPS,
+	DefaultValue: false,
+	Name:         "nohttps",
+	Deprecated:   "use --no-https",
+	Usage:        "use http instead of https for docker:// oras:// and library://<hostname>/... URIs",
+}
+
 // --passphrase
 var commonPromptForPassphraseFlag = cmdline.Flag{
 	ID:           "commonPromptForPassphraseFlag",
@@ -183,26 +265,6 @@ var commonForceFlag = cmdline.Flag{
 	EnvKeys:      []string{"FORCE"},
 }
 
-// --no-https
-var commonNoHTTPSFlag = cmdline.Flag{
-	ID:           "commonNoHTTPSFlag",
-	Value:        &noHTTPS,
-	DefaultValue: false,
-	Name:         "no-https",
-	Usage:        "use http instead of https for docker:// oras:// and library://<hostname>/... URIs",
-	EnvKeys:      []string{"NOHTTPS", "NO_HTTPS"},
-}
-
-// --nohttps (deprecated)
-var commonOldNoHTTPSFlag = cmdline.Flag{
-	ID:           "commonOldNoHTTPSFlag",
-	Value:        &noHTTPS,
-	DefaultValue: false,
-	Name:         "nohttps",
-	Deprecated:   "use --no-https",
-	Usage:        "use http instead of https for docker:// oras:// and library://<hostname>/... URIs",
-}
-
 // --tmpdir
 var commonTmpDirFlag = cmdline.Flag{
 	ID:           "commonTmpDirFlag",
@@ -214,15 +276,84 @@ var commonTmpDirFlag = cmdline.Flag{
 	EnvKeys:      []string{"TMPDIR"},
 }
 
-// -c|--config
-var singConfigFileFlag = cmdline.Flag{
-	ID:           "singConfigFileFlag",
-	Value:        &configurationFile,
-	DefaultValue: buildcfg.SINGULARITY_CONF_FILE,
-	Name:         "config",
-	ShortHand:    "c",
-	Usage:        "specify a configuration file (for root or unprivileged installation only)",
-	EnvKeys:      []string{"CONFIG_FILE"},
+// --oci
+var commonOCIFlag = cmdline.Flag{
+	ID:           "actionOCI",
+	Value:        &isOCI,
+	DefaultValue: false,
+	Name:         "oci",
+	Usage:        "Launch container with OCI runtime (experimental)",
+	EnvKeys:      []string{"OCI"},
+}
+
+// --no-oci
+var commonNoOCIFlag = cmdline.Flag{
+	ID:           "commonNoOCI",
+	Value:        &noOCI,
+	DefaultValue: false,
+	Name:         "no-oci",
+	Usage:        "Launch container with native runtime",
+	EnvKeys:      []string{"NO_OCI"},
+}
+
+// --keep-layers
+var commonKeepLayersFlag = cmdline.Flag{
+	ID:           "keepLayers",
+	Value:        &keepLayers,
+	DefaultValue: false,
+	Name:         "keep-layers",
+	Usage:        "Keep layers when creating an OCI-SIF. Do not squash to a single layer.",
+	EnvKeys:      []string{"KEEP_LAYERS"},
+}
+
+// --tmp-sandbox
+var actionTmpSandbox = cmdline.Flag{
+	ID:           "actionTmpSandbox",
+	Value:        &tmpSandbox,
+	DefaultValue: false,
+	Name:         "tmp-sandbox",
+	Usage:        "Forces unpacking of images into temporary sandbox dirs when a kernel or FUSE mount would otherwise be used.",
+	EnvKeys:      []string{"TMP_SANDBOX"},
+}
+
+// --no-tmp-sandbox
+var actionNoTmpSandbox = cmdline.Flag{
+	ID:           "actionNoTmpSandbox",
+	Value:        &noTmpSandbox,
+	DefaultValue: false,
+	Name:         "no-tmp-sandbox",
+	Usage:        "Prohibits unpacking of images into temporary sandbox dirs",
+	EnvKeys:      []string{"NO_TMP_SANDBOX"},
+}
+
+// --arch
+var commonArchFlag = cmdline.Flag{
+	ID:           "commonArchFlag",
+	Value:        &arch,
+	DefaultValue: "",
+	Name:         "arch",
+	Usage:        "architecture to use when pulling images",
+	EnvKeys:      []string{"PULL_ARCH", "ARCH"},
+}
+
+// --platform
+var commonPlatformFlag = cmdline.Flag{
+	ID:           "commonPlatformFlag",
+	Value:        &platform,
+	DefaultValue: "",
+	Name:         "platform",
+	Usage:        "platform (OS/Architecture/Variant) to use when pulling images",
+	EnvKeys:      []string{"PLATFORM"},
+}
+
+// --authfile
+var commonAuthFileFlag = cmdline.Flag{
+	ID:           "commonAuthFileFlag",
+	Value:        &reqAuthFile,
+	DefaultValue: "",
+	Name:         "authfile",
+	Usage:        "Docker-style authentication file to use for writing/reading OCI registry credentials",
+	EnvKeys:      []string{"AUTHFILE"},
 }
 
 func getCurrentUser() *user.User {
@@ -242,6 +373,8 @@ func setSylogMessageLevel() {
 
 	if debug {
 		level = 5
+		// Propagate debug flag to nested `singularity` calls.
+		os.Setenv("SINGULARITY_DEBUG", "1")
 	} else if verbose {
 		level = 4
 	} else if quiet {
@@ -262,25 +395,26 @@ func setSylogMessageLevel() {
 
 // handleRemoteConf will make sure your 'remote.yaml' config file
 // has the correct permission.
-func handleRemoteConf(remoteConfFile string) {
+func handleRemoteConf(remoteConfFile string) error {
 	// Only check the permission if it exists.
 	if fs.IsFile(remoteConfFile) {
 		sylog.Debugf("Ensuring file permission of 0600 on %s", remoteConfFile)
 		if err := fs.EnsureFileWithPermission(remoteConfFile, 0o600); err != nil {
-			sylog.Fatalf("Unable to correct the permission on %s: %s", remoteConfFile, err)
+			return fmt.Errorf("unable to correct the permission on %s: %w", remoteConfFile, err)
 		}
 	}
+	return nil
 }
 
 // handleConfDir tries to create the user's configuration directory and handles
 // messages and/or errors.
-func handleConfDir(confDir string) {
+func handleConfDir(confDir string) error {
 	if err := fs.Mkdir(confDir, 0o700); err != nil {
 		if os.IsExist(err) {
 			sylog.Debugf("%s already exists. Not creating.", confDir)
 			fi, err := os.Stat(confDir)
 			if err != nil {
-				sylog.Fatalf("Failed to retrieve information for %s: %s", confDir, err)
+				return fmt.Errorf("failed to retrieve information for %s: %s", confDir, err)
 			}
 			if fi.Mode().Perm() != 0o700 {
 				sylog.Debugf("Enforce permission 0700 on %s", confDir)
@@ -296,29 +430,62 @@ func handleConfDir(confDir string) {
 	} else {
 		sylog.Debugf("Created %s", confDir)
 	}
+	return nil
 }
 
-func persistentPreRun(*cobra.Command, []string) {
+func persistentPreRun(*cobra.Command, []string) error {
 	setSylogMessageLevel()
 	sylog.Debugf("Singularity version: %s", buildcfg.PACKAGE_VERSION)
 
 	if os.Geteuid() != 0 && buildcfg.SINGULARITY_SUID_INSTALL == 1 {
 		if configurationFile != singConfigFileFlag.DefaultValue {
-			sylog.Fatalf("--config requires to be root or an unprivileged installation")
+			return fmt.Errorf("--config requires to be root or an unprivileged installation")
 		}
 	}
 
 	sylog.Debugf("Parsing configuration file %s", configurationFile)
 	config, err := singularityconf.Parse(configurationFile)
 	if err != nil {
-		sylog.Fatalf("Couldn't not parse configuration file %s: %s", configurationFile, err)
+		return fmt.Errorf("couldn't parse configuration file %s: %s", configurationFile, err)
 	}
 	singularityconf.SetCurrentConfig(config)
 
+	// Honor 'oci mode' in singularity.conf, and allow negation with `--no-oci`.
+	if isOCI && noOCI {
+		return fmt.Errorf("--oci and --no-oci cannot be used together")
+	}
+	isOCI = isOCI || config.OCIMode
+	if noOCI {
+		isOCI = false
+	}
+
+	// --keep-layers is only valid in OCI mode, as native SIFs do not hold layers.
+	if keepLayers && !isOCI {
+		sylog.Fatalf("--keep-layers is only supported when creating OCI-SIF images (--oci mode)")
+	}
+
+	// Honor 'tmp sandbox' in singularity.conf, and allow negation with
+	// `--no-tmp-sandbox`.
+	canUseTmpSandbox = config.TmpSandboxAllowed
+	if noTmpSandbox {
+		canUseTmpSandbox = false
+	}
+
+	// If we need to enter a namespace (oci-mode) do the re-exec now, before any
+	// other handling happens.
+	if err := maybeReExec(); err != nil {
+		return err
+	}
+
 	// Handle the config dir (~/.singularity),
 	// then check the remove conf file permission.
-	handleConfDir(syfs.ConfigDir())
-	handleRemoteConf(syfs.RemoteConf())
+	if err := handleConfDir(syfs.ConfigDir()); err != nil {
+		return fmt.Errorf("while handling config dir: %w", err)
+	}
+	if err := handleRemoteConf(syfs.RemoteConf()); err != nil {
+		return fmt.Errorf("while handling remote config: %w", err)
+	}
+	return nil
 }
 
 // Init initializes and registers all singularity commands.
@@ -341,8 +508,16 @@ func Init(loadPlugins bool) {
 
 	// set persistent pre run function here to avoid initialization loop error
 	singularityCmd.PersistentPreRunE = func(cmd *cobra.Command, args []string) error {
-		persistentPreRun(cmd, args)
-		return cmdManager.UpdateCmdFlagFromEnv(cmd, envPrefix)
+		if err := cmdManager.UpdateCmdFlagFromEnv(singularityCmd, envPrefix); err != nil {
+			sylog.Fatalf("While parsing global environment variables: %s", err)
+		}
+		if err := cmdManager.UpdateCmdFlagFromEnv(cmd, envPrefix); err != nil {
+			sylog.Fatalf("While parsing environment variables: %s", err)
+		}
+		if err := persistentPreRun(cmd, args); err != nil {
+			sylog.Fatalf("While initializing: %s", err)
+		}
+		return nil
 	}
 
 	cmdManager.RegisterFlagForCmd(&singDebugFlag, singularityCmd)
@@ -360,6 +535,7 @@ func Init(loadPlugins bool) {
 	}
 
 	// load plugins and register commands/flags if any
+	//nolint:forcetypeassert
 	if loadPlugins {
 		callbackType := (clicallback.Command)(nil)
 		callbacks, err := plugin.LoadCallbacks(callbackType)
@@ -581,7 +757,7 @@ func singularityExec(image string, args []string) (string, error) {
 }
 
 // CheckRoot ensures that a command is executed with root privileges.
-func CheckRoot(cmd *cobra.Command, args []string) {
+func CheckRoot(cmd *cobra.Command, _ []string) {
 	if os.Geteuid() != 0 {
 		sylog.Fatalf("%q command requires root privileges", cmd.CommandPath())
 	}
@@ -589,7 +765,7 @@ func CheckRoot(cmd *cobra.Command, args []string) {
 
 // CheckRootOrUnpriv ensures that a command is executed with root
 // privileges or that Singularity is installed unprivileged.
-func CheckRootOrUnpriv(cmd *cobra.Command, args []string) {
+func CheckRootOrUnpriv(cmd *cobra.Command, _ []string) {
 	if os.Geteuid() != 0 && buildcfg.SINGULARITY_SUID_INSTALL == 1 {
 		sylog.Fatalf("%q command requires root privileges or an unprivileged installation", cmd.CommandPath())
 	}
@@ -657,3 +833,45 @@ func getBuilderClientConfig(uri string) (baseURI, authToken string, err error) {
 
 	return currentRemoteEndpoint.BuilderClientConfig(uri)
 }
+
+func maybeReExec() error {
+	sylog.Debugf("Checking whether to re-exec")
+	// The OCI runtime must always be launched where the effective uid/gid is 0 (root or fake-root).
+	if isOCI && !rootless.InNS() {
+		// If we need to, enter a new cgroup now, to workaround an issue with crun container cgroup creation (#1538).
+		if err := ocilauncher.CrunNestCgroup(); err != nil {
+			return fmt.Errorf("while applying crun cgroup workaround: %w", err)
+		}
+		// If we are root already, run the launcher in a new mount namespace only.
+		if os.Geteuid() == 0 {
+			return rootless.RunInMountNS(os.Args[1:])
+		}
+		// If we are not root, re-exec in a root-mapped user namespace and mount namespace.
+		return rootless.ExecWithFakeroot(os.Args[1:])
+	}
+	return nil
+}
+
+// getOCIPlatform returns the appropriate OCI platform to use according to `--arch` and `--platform`
+func getOCIPlatform() ggcrv1.Platform {
+	var (
+		p   *ggcrv1.Platform
+		err error
+	)
+	if arch != "" && platform != "" {
+		err = fmt.Errorf("--arch and --platform cannot be used together")
+	}
+	if arch == "" && platform == "" {
+		p, err = ociplatform.DefaultPlatform()
+	}
+	if arch != "" {
+		p, err = ociplatform.PlatformFromArch(arch)
+	}
+	if platform != "" {
+		p, err = ociplatform.PlatformFromString(platform)
+	}
+	if err != nil {
+		sylog.Fatalf("%v", err)
+	}
+	return *p
+}
