@@ -13,7 +13,6 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"net"
 	"os"
 	"os/exec"
@@ -30,19 +29,19 @@ import (
 	"unsafe"
 
 	specs "github.com/opencontainers/runtime-spec/specs-go"
-	"github.com/sylabs/singularity/internal/pkg/instance"
-	"github.com/sylabs/singularity/internal/pkg/plugin"
-	"github.com/sylabs/singularity/internal/pkg/security"
-	"github.com/sylabs/singularity/internal/pkg/util/env"
-	"github.com/sylabs/singularity/internal/pkg/util/fs/files"
-	"github.com/sylabs/singularity/internal/pkg/util/machine"
-	"github.com/sylabs/singularity/internal/pkg/util/shell"
-	"github.com/sylabs/singularity/internal/pkg/util/shell/interpreter"
-	"github.com/sylabs/singularity/internal/pkg/util/user"
-	singularitycallback "github.com/sylabs/singularity/pkg/plugin/callback/runtime/engine/singularity"
-	singularityConfig "github.com/sylabs/singularity/pkg/runtime/engine/singularity/config"
-	"github.com/sylabs/singularity/pkg/sylog"
-	"github.com/sylabs/singularity/pkg/util/rlimit"
+	"github.com/sylabs/singularity/v4/internal/pkg/instance"
+	"github.com/sylabs/singularity/v4/internal/pkg/plugin"
+	"github.com/sylabs/singularity/v4/internal/pkg/security"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/env"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/fs/files"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/machine"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/shell"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/shell/interpreter"
+	"github.com/sylabs/singularity/v4/internal/pkg/util/user"
+	singularitycallback "github.com/sylabs/singularity/v4/pkg/plugin/callback/runtime/engine/singularity"
+	singularityConfig "github.com/sylabs/singularity/v4/pkg/runtime/engine/singularity/config"
+	"github.com/sylabs/singularity/v4/pkg/sylog"
+	"github.com/sylabs/singularity/v4/pkg/util/rlimit"
 	"golang.org/x/sys/unix"
 	"golang.org/x/term"
 	"mvdan.cc/sh/v3/interp"
@@ -56,6 +55,7 @@ const defaultShell = "/bin/sh"
 // No additional privileges can be gained during this call (unless container
 // is executed as root intentionally) as starter will set uid/euid/suid
 // to the targetUID (PrepareConfig will set it by calling starter.Config.SetTargetUID).
+//
 //nolint:maintidx
 func (e *EngineOperations) StartProcess(masterConn net.Conn) error {
 	// Manage all signals.
@@ -202,6 +202,7 @@ func (e *EngineOperations) StartProcess(masterConn net.Conn) error {
 		}
 		if err := cmd.Start(); err != nil {
 			if e, ok := err.(*os.PathError); ok {
+				//nolint:forcetypeassert
 				if e.Err.(syscall.Errno) == syscall.ENOEXEC && args[0] != defaultShell {
 					args = append([]string{defaultShell}, args...)
 					goto cmdexec
@@ -263,6 +264,7 @@ func (e *EngineOperations) StartProcess(masterConn net.Conn) error {
 				// https://github.com/golang/go/issues/24543.
 				break
 			default:
+				//nolint:forcetypeassert
 				signal := s.(syscall.Signal)
 				// EPERM and EINVAL are deliberately ignored because they can't be
 				// returned in this context, this process is PID 1, so it has the
@@ -291,6 +293,7 @@ func (e *EngineOperations) StartProcess(masterConn net.Conn) error {
 			} else if e, ok := err.(*os.SyscallError); ok {
 				// handle possible race with Wait4 call above by ignoring ECHILD
 				// error because child process was already caught
+				//nolint:forcetypeassert
 				if e.Err.(syscall.Errno) != syscall.ECHILD {
 					sylog.Fatalf("error while waiting container process: %s", e.Error())
 				}
@@ -321,7 +324,7 @@ func (e *EngineOperations) StartProcess(masterConn net.Conn) error {
 // and thus no additional privileges can be gained.
 //
 // Here, however, singularity engine does not escalate privileges.
-func (e *EngineOperations) PostStartProcess(ctx context.Context, pid int) error {
+func (e *EngineOperations) PostStartProcess(_ context.Context, pid int) error {
 	sylog.Debugf("Post start process")
 
 	callbackType := (singularitycallback.PostStartProcess)(nil)
@@ -330,6 +333,7 @@ func (e *EngineOperations) PostStartProcess(ctx context.Context, pid int) error
 		return fmt.Errorf("while loading plugins callbacks '%T': %s", callbackType, err)
 	}
 	for _, cb := range callbacks {
+		//nolint:forcetypeassert
 		if err := cb.(singularitycallback.PostStartProcess)(e.CommonConfig, pid); err != nil {
 			return err
 		}
@@ -670,7 +674,7 @@ func injectEnvHandler(senv map[string]string, noEval bool) interpreter.OpenHandl
 	}
 }
 
-func runtimeVarsHandler(senv map[string]string) interpreter.OpenHandler {
+func runtimeVarsHandler() interpreter.OpenHandler {
 	var once sync.Once
 
 	return func(path string, _ int, _ os.FileMode) (io.ReadWriteCloser, error) {
@@ -685,7 +689,7 @@ func runtimeVarsHandler(senv map[string]string) interpreter.OpenHandler {
 }
 
 // sylogBuiltin allows to use sylog logger from shell script.
-func sylogBuiltin(ctx context.Context, argv []string) error {
+func sylogBuiltin(_ context.Context, argv []string) error {
 	if len(argv) < 2 {
 		return fmt.Errorf("sylog builtin requires two arguments")
 	}
@@ -705,7 +709,7 @@ func sylogBuiltin(ctx context.Context, argv []string) error {
 }
 
 // getAllEnvBuiltin display all exported variables in the form KEY=VALUE.
-func getAllEnvBuiltin(shell *interpreter.Shell) interpreter.ShellBuiltin {
+func getAllEnvBuiltin() interpreter.ShellBuiltin {
 	return func(ctx context.Context, argv []string) error {
 		hc := interp.HandlerCtx(ctx)
 
@@ -739,7 +743,7 @@ func getAllEnvBuiltin(shell *interpreter.Shell) interpreter.ShellBuiltin {
 
 // fixPathBuiltin takes the current path value to fix it by injecting
 // missing default path and returns value on shell interpreter output.
-func fixPathBuiltin(ctx context.Context, argv []string) error {
+func fixPathBuiltin(ctx context.Context, _ []string) error {
 	hc := interp.HandlerCtx(ctx)
 
 	currentPath := filepath.SplitList(hc.Env.Get("PATH").String())
@@ -765,7 +769,7 @@ func fixPathBuiltin(ctx context.Context, argv []string) error {
 
 // hashBuiltin is a noop function for hash bash builtin, since we don't
 // store resolved path in a hash table, there is nothing to do.
-func hashBuiltin(ctx context.Context, argv []string) error {
+func hashBuiltin(context.Context, []string) error {
 	return nil
 }
 
@@ -815,10 +819,10 @@ func runActionScript(engineConfig *singularityConfig.EngineConfig) ([]string, []
 	senv := engineConfig.GetSingularityEnv()
 	shell.RegisterOpenHandler("/.inject-singularity-env.sh", injectEnvHandler(senv, engineConfig.GetNoEval()))
 
-	shell.RegisterOpenHandler("/.singularity.d/env/99-runtimevars.sh", runtimeVarsHandler(senv))
+	shell.RegisterOpenHandler("/.singularity.d/env/99-runtimevars.sh", runtimeVarsHandler())
 
 	// register few builtin
-	shell.RegisterShellBuiltin("getallenv", getAllEnvBuiltin(shell))
+	shell.RegisterShellBuiltin("getallenv", getAllEnvBuiltin())
 	shell.RegisterShellBuiltin("sylog", sylogBuiltin)
 	shell.RegisterShellBuiltin("fixpath", fixPathBuiltin)
 	shell.RegisterShellBuiltin("hash", hashBuiltin)
@@ -830,7 +834,7 @@ func runActionScript(engineConfig *singularityConfig.EngineConfig) ([]string, []
 	// function to execute the command
 	shell.RegisterShellBuiltin("exec", execBuiltin)
 
-	err = shell.Run()
+	err = shell.Run(context.TODO())
 	if err != nil {
 		if shell.Status() != 0 {
 			os.Exit(int(shell.Status()))
@@ -848,7 +852,7 @@ func runActionScript(engineConfig *singularityConfig.EngineConfig) ([]string, []
 				return nil, nil, err
 			}
 			interp.RegisterShellBuiltin("exec", execBuiltin)
-			err = interp.Run()
+			err = interp.Run(context.TODO())
 			if err != nil {
 				if interp.Status() != 0 {
 					os.Exit(int(interp.Status()))
@@ -864,7 +868,7 @@ func runActionScript(engineConfig *singularityConfig.EngineConfig) ([]string, []
 // getDockerRunscript returns the content as a reader of
 // the default runscript set for docker images if any.
 func getDockerRunscript(path string) (io.Reader, error) {
-	r, err := ioutil.ReadFile(path)
+	r, err := os.ReadFile(path)
 	if err != nil {
 		return nil, fmt.Errorf("while reading %s: %s", path, err)
 	}
